#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "tsmforth.h"
#include "strhash.h"
#include "debug.h"

static int tsm_handle_call(tsm_ctx *ctx)
{
  tsm_data_e type;
  int *data;
  long int entry;

  if(tsm_stack_getpop(ctx->stack,&type,(void *)&data)<0)
    return(-1);

  if(type!=TFD_INT)
    return(-1);

  entry=(long int)hash_lookup_key(ctx->code->entry,*data);
  if(entry==0) return(-1);

  DEBUG_PRINT(("	CALL %s\n",hash_get_str(ctx->code->entry,*(unsigned *)data)));
  ctx->ip[++(ctx->depth)]=entry;
  return(0);
}

static int tsm_handle_nop(tsm_ctx *ctx) { return(0); }

static int tsm_handle_pop(tsm_ctx *ctx)
{
  return(tsm_stack_pop(ctx->stack));
}

static int tsm_handle_invalid(tsm_ctx *ctx)
{
  fprintf(stderr,"Invalid Instruction\n");
  return(-1);
}

static int tsm_handle_push(tsm_ctx *ctx)
{
  return(tsm_stack_push(ctx->stack,
                       *((tsm_data_e *)ctx->ip[ctx->depth]),
                       (void *)(ctx->ip[ctx->depth]+sizeof(tsm_data_e))));
}

static int tsm_handle_if(tsm_ctx *ctx)
{
  int *off,*value;

  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&off)<0)
    return(-1);

  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&value)<0)
    return(-1);

  if(value[0])
    DEBUG_PRINT(("\tEvaluates true, not skipping\n"));
  else
  {
    ctx->ip[ctx->depth]+=off[0];
    DEBUG_PRINT(("\tEvaluates false, jumping to %ld\n",ctx->ip[ctx->depth]));
  }

  return(0);
}

static int tsm_handle_ifn(tsm_ctx *ctx)
{
  int *off,*value;

  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&off)<0)
    return(-1);

  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&value)<0)
    return(-1);

  if(value[0])
  {
    ctx->ip[ctx->depth]+=off[0];
    DEBUG_PRINT(("\tEvaluates true, jumping to %ld\n",ctx->ip[ctx->depth]));
  }
  else
    DEBUG_PRINT(("\tEvaluates false, not skipping\n"));

  return(0);
}

static int tsm_handle_dup(tsm_ctx *ctx)
{
  tsm_data_e type;
  void *data;
  if(tsm_stack_peek(ctx->stack,&type,&data)<0)
    return(-1);
  return(tsm_stack_push(ctx->stack,type,data));
}

static int tsm_handle_tell(tsm_ctx *ctx)
{
  tsm_data_e type;
  void *data;
  if(tsm_stack_getpop(ctx->stack,&type,&data)<0)
    return(-1);

  switch(type)
  {
  case TFD_INT:
    printf("%d",*(const int *)data);
    break;
  case TFD_STRING:
    printf("%s",(const char *)data);
    break;
  default:
    if((type<0)||(type>=type_max))
      return(-1);
    if(vhandlers[type].print == NULL)
      return(-1);

    vhandlers[type].print(data);
    break;
  }
  return(0);
}

static int tsm_handle_getl(tsm_ctx *ctx)
{
  char buf[64];
  int *data;
  tsm_codedata *tfd;

  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&data)<0)
    return(-1);

  tsm_get_var_key(buf,data[0]);
  DEBUG_PRINT(("\tVariable key %s\n",buf));

  tfd=(tsm_codedata *)hash_lookup_str(ctx->locals,buf);
  if(tfd==NULL)
  {
    tsm_debug("\tCouldn't look up key(%x)\n",data[0]);
    return(-1);
  }

  tsm_stack_pop(ctx->stack);

  return(tsm_stack_push(ctx->stack,tfd->type,tfd->value.strvalue));
}

static int tsm_handle_setl(tsm_ctx *ctx)
{
  char buf[64];
  int size=sizeof(tsm_data_e);
  tsm_data_e type;
  const int *data;
  void *stdata;
  tsm_codedata *tfd;

  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&data)<0)
  {
    tsm_debug("	Stack underflow or bad type\n");
    return(-1);
  }

  tsm_get_var_key(buf,data[0]);
  DEBUG_PRINT(("\tVariable key %s\n",buf));

  tfd=(tsm_codedata *)hash_remove_str(ctx->locals,buf);
  if(tfd!=NULL) free(tfd);

  if(tsm_stack_getpop(ctx->stack,&type,&stdata)<0)
  {
    tsm_debug("	Underflow\n");
    return(-1);
  }

  size=tsm_var_size(type,stdata);
  tfd=malloc(size+sizeof(tsm_data_e));
  tfd->type=type;
  memcpy(tfd->value.strvalue,stdata,size);
  if(hash_add_str(ctx->locals,buf,tfd)<0)
  {
    free(tfd);
    return(-1);
  }
  return(0);
}

static int tsm_handle_rand(tsm_ctx *ctx)
{
  return(tsm_stack_push_int(ctx->stack,rand()));
}

static int tsm_handle_add(tsm_ctx *ctx)
{
  tsm_data_e type;
  void *d1,*d2;

  if(tsm_stack_getpop(ctx->stack,&type,&d1)<0)
  {
    tsm_debug("	Stack underflow\n");
    return(-1);
  }
  if(tsm_stack_getpop_require(ctx->stack,type,&d2)<0)
  {
    tsm_debug("	Error retrieving second operand\n");
    return(-1);
  }

  switch(type)
  {  
  int retval;
  char *str;
  case TFD_INT:
    retval=*(int *)d1;
    retval+=*(int *)d2;
    return(tsm_stack_push_int(ctx->stack,retval));

  case TFD_STRING:
    str=malloc(strlen(d1) + strlen(d2) + 1);
    if(str==NULL) return(-1);
    strcpy(str,d2);
    strcat(str,d1);
    retval=tsm_stack_push_string(ctx->stack,str);
    free(str);
    return(retval);
  default:
    tsm_debug("	Invalid type\n");
    return(-1);
  }
}

static int tsm_handle_sub(tsm_ctx *ctx)
{
  tsm_data_e type=TFD_INT;
  int *d1,*d2;

  if(tsm_stack_getpop_require(ctx->stack,type,(void **)&d1)<0)
    return(-1);
  if(tsm_stack_getpop_require(ctx->stack,type,(void **)&d2)<0)
    return(-1);
  return(tsm_stack_push_int(ctx->stack,d2[0]-d1[0]));
}

static int tsm_handle_mul(tsm_ctx *ctx)
{
  tsm_data_e type=TFD_INT;
  int *d1,*d2;

  if(tsm_stack_getpop_require(ctx->stack,type,(void **)&d1)<0)
    return(-1);
  if(tsm_stack_getpop_require(ctx->stack,type,(void **)&d2)<0)
    return(-1);
  return(tsm_stack_push_int(ctx->stack,d1[0]*d2[0]));
}

static int tsm_handle_mod(tsm_ctx *ctx)
{
  int *d1,*d2;

  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&d1)<0)
    return(-1);
  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&d2)<0)
    return(-1);
  if(d1[0]==0)
    return(-1);
  return(tsm_stack_push_int(ctx->stack,d2[0]%d1[0]));
}

static int tsm_handle_div(tsm_ctx *ctx)
{
  tsm_data_e type=TFD_INT;
  int *d1,*d2;

  if(tsm_stack_getpop_require(ctx->stack,type,(void **)&d1)<0)
    return(-1);
  if(tsm_stack_getpop_require(ctx->stack,type,(void **)&d2)<0)
    return(-1);
  if(d1[0]==0)
    return(-1);
  return(tsm_stack_push_int(ctx->stack,d2[0]/d1[0]));
}

static int tsm_handle_tostr(tsm_ctx *ctx)
{
  char buf[512];
  tsm_data_e type;
  void *data;

  if(tsm_stack_getpop(ctx->stack,&type,&data)<0)    return(-1);

  switch(type)
  {
  case TFD_INT:
    sprintf(buf,"%d",*(const int *)data);
    return(tsm_stack_push_string(ctx->stack,buf));

  case TFD_STRING:
    sprintf(buf,"%s",(const char *)data);
    return(tsm_stack_push_string(ctx->stack,buf));

  default:
    if((type<0)||(type>=type_max))                   return(-1);
    if(vhandlers[type].tostr == NULL)                return(-1);
    if(vhandlers[type].tostr(buf,data)<0)            return(-1);
    return(tsm_stack_push_string(ctx->stack,buf));
  }

}

static int tsm_handle_not(tsm_ctx *ctx)
{
  int *off,*value;

  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&value)<0)
    return(-1);

  return(tsm_stack_push_int(ctx->stack,!(*value)));
}

static int tsm_handle_eq(tsm_ctx *ctx)
{
  int output;
  tsm_data_e type;
  void *d1,*d2;

  if(tsm_stack_getpop(ctx->stack,&type,&d1)<0)
    return(-1);
  if(tsm_stack_getpop_require(ctx->stack,type,&d2)<0)
    return(-1);

  switch(type)
  {
  case TFD_INT:
    output=(*(int *)d1)==(*(int *)d2);
    break;
  case TFD_STRING:
    output=(strcmp((const char *)d1,(const char *)d2)==0);
    break;
  default:
    return(-1);
  }

  return(tsm_stack_push_int(ctx->stack,output));
}

static int tsm_handle_bor(tsm_ctx *ctx)
{
  int *d1,*d2;

  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&d1)<0)
    return(-1);
  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&d2)<0)
    return(-1);
  return(tsm_stack_push_int(ctx->stack,d2[0]|d1[0]));
}

static int tsm_handle_band(tsm_ctx *ctx)
{
  int *d1,*d2;

  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&d1)<0)
    return(-1);
  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&d2)<0)
    return(-1);
  return(tsm_stack_push_int(ctx->stack,d2[0]&d1[0]));
}

static int tsm_handle_bxor(tsm_ctx *ctx)
{
  int *d1,*d2;

  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&d1)<0)
    return(-1);
  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&d2)<0)
    return(-1);
  return(tsm_stack_push_int(ctx->stack,d2[0]^d1[0]));
}

static int tsm_handle_inv(tsm_ctx *ctx)
{
  int *d1;

  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&d1)<0)
    return(-1);
  return(tsm_stack_push_int(ctx->stack,~d1[0]));
}

static int tsm_handle_neg(tsm_ctx *ctx)
{
  int *d1;

  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&d1)<0)
    return(-1);
  return(tsm_stack_push_int(ctx->stack,-d1[0]));
}

static int tsm_handle_land(tsm_ctx *ctx)
{
  int *d1,*d2;

  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&d1)<0)
    return(-1);
  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&d2)<0)
    return(-1);
  return(tsm_stack_push_int(ctx->stack,d2[0]&&d1[0]));
}

static int tsm_handle_lor(tsm_ctx *ctx)
{
  int *d1,*d2;

  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&d1)<0)
    return(-1);
  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&d2)<0)
    return(-1);
  return(tsm_stack_push_int(ctx->stack,d2[0]||d1[0]));
}

static int tsm_handle_depth(tsm_ctx *ctx)
{
  return(tsm_stack_push_int(ctx->stack,ctx->stack->depth));
}

static int tsm_handle_jump(tsm_ctx *ctx)
{
  int *off;
  if(tsm_stack_getpop_require(ctx->stack,TFD_INT,(void **)&off)<0)
    return(-1);

  ctx->ip[ctx->depth]+=off[0];
  DEBUG_PRINT(("\tHard-jump to %lx\n",ctx->ip[ctx->depth]));

  return(0);
}

static int tsm_handle_randomize(tsm_ctx *ctx)
{
  srand((unsigned int)time(NULL));
  return(0);
}

const tsm_handler tsm_instr[]={
  /* DO NOT CHANGE THE ORDER OF THESE FUNDAMENTAL OPS! */
  { tsm_handle_invalid,"????"}
, { tsm_handle_if   ,IF_STR }
, { tsm_handle_ifn  ,"IFN"  }
, { tsm_handle_jump ,"JUMP" }
, { tsm_handle_call ,"CALL" }
, { tsm_handle_nop  ,";"    }
, { tsm_handle_setl ,"!"    }
, { tsm_handle_getl ,"@"    }
  /* End of fundamental ops */

  /* Put any other ops here in whatever order. */
, { tsm_handle_pop  ,"POP"  } // POP   -- pop one element off the stack        
, { tsm_handle_dup  ,"DUP"  } // DUP   -- duplicate top element of stack       
, { tsm_handle_tell ,"TELL" } // TELL  -- pop/print top element of stack       
, { tsm_handle_rand ,"RAND" } // RAND  -- push random # onto stack             
, { tsm_handle_tostr,"TOSTR"} // TOSTR -- pop top element, convert, push string
, { tsm_handle_not  ,"NOT"  } // NOT   -- C-style integer inversion            
, { tsm_handle_depth,"DEPTH"} // DEPTH -- Push stack-depth as int              
, { tsm_handle_neg  ,"NEG"  } // NEG   -- pop a, push -a                       
, { tsm_handle_randomize,"RANDOMIZE"} // Guess
, { tsm_handle_add  ,"+"    } // +     -- Add for int, strcat for string       
, { tsm_handle_sub  ,"-"    } // -     -- subtract integer                     
, { tsm_handle_mod  ,"%"    } // %     -- integer modulous                     
, { tsm_handle_mul  ,"*"    } // *     -- integer multiply                     
, { tsm_handle_div  ,"/"    } // /     -- integer divide                       
, { tsm_handle_eq   ,"="    } // =     -- pop a, pop b, push a == b            
, { tsm_handle_bor  ,"|"    } // |     -- pop a, pop b, push a | b             
, { tsm_handle_band ,"&"    } // &     -- pop a, pop b, push a & b             
, { tsm_handle_bxor ,"^"    } // ^     -- pop a, pop b, push a ^ b             
, { tsm_handle_inv  ,"~"    } // ~     -- pop a, push ~a                       
, { tsm_handle_land ,"&&"   } // &&    -- pop a, pop b, push a && b            
, { tsm_handle_lor  ,"||"   } // ||    -- pop a, pop b, push a || b            

  /* This entry means 'end of list', do not leave it out */
, { NULL           ,NULL   } };

const int TSM_LAST=sizeof(tsm_instr)/sizeof(tsm_handler);
