/**
 * strhash.h -- character-string hashtables with linear probing
 *
 * Implements a simple hashtable which you can insert and delete
 * elements in.
 */
#ifndef __STRHASH_H__
#define __STRHASH_H__

#include <string.h>

/**
 * The structure the hashtable is founded on -- has an input string
 * and an anonymous pointer output.  The pointer must be non-NULL
 * to be valid.
 */
typedef struct strhash_entry
{
  const char *key;
  void *output;
} strhash_entry;

/**
 * Hash table.  Contains table and size.
 */
typedef struct strtable
{
  strhash_entry *table;
  int size;
} strtable;

#ifdef __cplusplus
extern "C" {
#endif
/* Hash function.  Returns a 'fingerprint' for a string. */
unsigned int str_hash(const char *str);

/* Dynamically allocates a hash table. */
strtable   *hash_create(int size);
/* Frees a dynamically-allocated hashtable. */
void        hash_free(strtable *st);
/* Adds a key-value pair to a dynamic-data hashtable.  The value of 
 * 'str' is copied.
 */
int         hash_add_str(strtable *st, const char *str, void *value);
/* Deletes a key-value pair from the table, returning the value.  
 *  Returns NULL of absent.
 */
void       *hash_remove_str(strtable *st, const char *str);
/* Removes a key-value pair from the hashtable, located by integer key.
 * Note that this integer key is NOT the index, it is the value of
 * str_hash(key_str).
 */
void       *hash_remove_key(strtable *st, unsigned int k);
/* Finds and returns the value for a given keystring, or NULL if absent */
void       *hash_lookup_str(strtable *st, const char *str);
/* Finds and returns the value for a given key, or NULL if absent */
void       *hash_lookup_key(strtable *st, unsigned int k);
/* Finds and returns the string for a given key, or NULL if absent */
const char *hash_get_str(strtable *st, unsigned int key);
/* Prints debugging output of hash table values */
void        hash_dump(strtable *st);

/* static-data hashtable functions */

/**
 * Macros for dealing with static-data hashtables.  These are allocated
 * completely at compile-time.
 */

/* Declares a static-data hashtable of fixed size. */
#define STATIC_HASHTABLE(name,size) strhash_entry name ## _entries[size];  strtable name={name ## _entries,size};
/* Initializes the hashtable at runtime.  Must only be called once, 
 *before table is used.
 */
#define STATIC_HASHINIT(name) memset(name ## _entries, 0, sizeof(name ## _entries));
/* Adds a key-value pair to a static-data hashtable.  The value of str 
 * is NOT duplicated, and must not change.
 */
int         hash_add_str_static(strtable *st, const char *str, void *value);
/* Removes a keystring-value pair from the static data hashtable, 
 * returning the value, or NULL if absent.
 */
void       *hash_remove_str_static(strtable *st, const char *str);
/* Removes a key-value pair from the static data hashtable by 
 * key-integer, returning the value, or NULL if absent
 */
void       *hash_remove_key_static(strtable *st, unsigned int k);

/* Experimental macros for iterating through hashtables */

#define HASH_END(st) ((st)->size)
#define HASH_KEYSTR(st,i) ((st)->table[i].key)
#define HASH_NEXT(st,i) do { (i)++; } while(((i)!=HASH_END(st)) && (HASH_KEYSTR(st,i)==NULL));
#define HASH_BEGIN(st,i) { (i)=-1; HASH_NEXT(st,i); }

#ifdef __cpluslpus
}
#endif

#endif/*__STRHASH_H__*/
