#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tsmforth.h"
#include "strhash.h"
#include "debug.h"

#define ALLOC(type,n) (type *)malloc(sizeof(type)*(n))
//#define PTR_EXTRACT(type,ptr)  (*((type *)(ptr)))

#define CUR_POS(ctx)           ((ctx)->ip[(ctx)->depth])
#define CUR_OFFSET(ctx)        ((ctx)->code->mmap)
#define CUR_DATA(ctx)          (CUR_POS(ctx) + CUR_OFFSET(ctx))
#define CUR_DATA_PTR(ctx,type) ((type *)CUR_DATA(ctx))
#define CUR_DATA_INC(ctx,size) CUR_POS(ctx) = align_up(CUR_POS(ctx) + (size)+sizeof(tsm_data_e));
//#define CUR_DATA_INCTYPE(ctx,type) CUR_DATA_INC(ctx,sizeof(type))

int tsm_execute(tsm_ctx *ctx);

#ifdef TSMFORTH_TEST
int main(int argc, char *argv[])
{
  int n;
  tsm_ctx  *ctx;
  tsm_code *tfc;

  if(argc<2)
  {
    tsm_debug("Syntax:  tsmforth file1.tsm [file2 ...] [ -- bottom ... top ]\n");
    return(1);
  }

  tfc=code_create(1024*1024*256);

  for(n=1; (n<argc)&&(strcmp("--",argv[n])!=0); n++)
  {

    if(forth_parse(tfc,argv[n])==NULL)
    {
      tsm_debug("Couldn't compile %s\n",argv[n]);
      code_free(tfc);
      return(1);
    }

    DEBUG_PRINT(("Compiled %s\n",argv[n]));
  }

  ctx=ctx_create(tfc,"main");

  n++;
  while(n<argc)
  {
    DEBUG_PRINT(("--> %s\n",argv[n]));
    if(tsm_stack_push_string(ctx->stack,argv[n])<0)
    {
      tsm_debug("Couldn't push arg onto stack\n");
      goto TEST_END;
    }

    n++;
  }

  DEBUG_PRINT(("Setting up entry point at %d\n",ctx->ip[0]));

  DEBUG_PRINT(("Beginning execution\n"));
  if(tsm_execute(ctx)<0)
    tsm_debug("Error in execution\n");

  DEBUG_PRINT(("Execution completed\n"));

TEST_END:
  ctx_free(ctx);
  code_free(tfc);
  fflush(stdout);
  return(0);
}
#endif/*TSMFORTH_TEST*/

int tsm_execute(tsm_ctx *ctx)
{
  int iter=0,running=1;

  while(running)
  {
    const tsm_codedata *cd=CUR_DATA_PTR(ctx,tsm_codedata);

#ifdef DEBUG

    DEBUG_PRINT(("%04d] 0x%08x ",iter++,ctx->ip[ctx->depth]));
    tsm_stack_print(ctx->stack);

#endif/*DEBUG*/

    switch(cd->type)
    {
    tsm_inst_e inst;
    default:
      tsm_debug("Unknown type '%d'\n",cd->type);
       return(-1);

    case TFD_EOF:
      tsm_debug("Unexpected EOF\n");
      return(-1);

    case TFD_INT:
      DEBUG_PRINT(("PUSH INT %d\n",cd->value.intvalue));
      if(tsm_stack_push_int(ctx->stack,cd->value.intvalue)<0)
        return(-1);

      CUR_DATA_INC(ctx,sizeof(int));
      break;

    case TFD_STRING:
      DEBUG_PRINT(("PUSH STR \"%s\"\n",cd->value.strvalue));
      if(tsm_stack_push_string(ctx->stack,cd->value.strvalue)<0)
        return(-1);

      CUR_DATA_INC(ctx,strlen(cd->value.strvalue)+1);
      break;

    case TFD_INSTRUCTION:
      inst=cd->value.instrvalue;
      CUR_DATA_INC(ctx,sizeof(tsm_inst_e));

      if((inst<0)||(inst>=TSM_LAST))
        return(-1);

      DEBUG_PRINT(("\t%s\n",tsm_instr[inst].token));

      if(inst == TSM_RET)
      {
        if(ctx->depth == 0)
        {
          DEBUG_PRINT(("\tReturning to operating system\n"));
          return(0);
        }

        DEBUG_PRINT(("\tRET\n"));
        ctx->depth--;
      }
      else if(tsm_instr[inst].handler(ctx)<0)
      {
        tsm_debug("  Invalid operation\n");
        return(-1);
      }

      break;
    }
  }
}

unsigned long code_getentry(tsm_code *tfc, const char *str)
{
  return((unsigned long)hash_lookup_str(tfc->entry,str));
}

tsm_code *code_create(int size)
{
  tsm_code *tfc;
//  fprintf(stderr,"code_create(%d)\n",size);
  if(size<=0) return(NULL);

  tfc=ALLOC(tsm_code,1);
  if(tfc==NULL) goto TFC_ERROR;

  memset(tfc,0,sizeof(tsm_code));

  tfc->mmap=ALLOC(unsigned char,size);
  if(tfc->mmap==NULL)
    goto TFC_ERROR;


  *(tsm_data_e *)(tfc->mmap) = TFD_EOF;

  tfc->mmap_pos=align_up(sizeof(tsm_data_e)); /* cannot be 0 */
  tfc->mmap_end=size;
  tfc->entry=hash_create(ENTRY_HASH);
  tfc->aux=NULL;

  if(tfc->entry==NULL)
    goto TFC_ERROR;

  return(tfc);

TFC_ERROR:
  if(tfc!=NULL)
    code_free(tfc);
  return(NULL);
}

void code_free(tsm_code *tfc)
{
  if(tfc!=NULL)
  {
    if(tfc->entry!=NULL) hash_free(tfc->entry);
    if(tfc->mmap !=NULL) free(tfc->mmap);
//    if(tfc->aux  !=NULL) free(tfc->aux);
    free(tfc);
  }
}

int code_addentry(tsm_code *tfc, const char *str)
{
//  fprintf(stderr,"code_addentry(%p,\"%s\")\n",tfc,str);
  if(tfc->mmap_pos >= tfc->mmap_end)
  {
    tsm_debug("\tMemory map full\n");
    return(-1);
  }

  if(hash_add_str(tfc->entry,str,(void *)tfc->mmap_pos)<0)
  {
    tsm_debug("\tFailed to add entry point -- hash collision or table full\n");
    return(-1);
  }
//  fprintf(stderr,"\tAdded entry point 0x%04x\n",(unsigned int)tfc->mmap_pos);
  return(0);
}

int code_addint(tsm_code *tfc, int val)
{
//  fprintf(stderr,"code_addint(%p,\"%d\")\n",tfc,val);

  if((tfc->mmap_pos+(sizeof(int)*2)) > tfc->mmap_end)
  {
    tsm_debug("\tCode seg full\n");
    return(-1);
  }

  {
    int *pos=(int *)(tfc->mmap + tfc->mmap_pos);
    pos[0]=TFD_INT;
    pos[1]=val;
    tfc->mmap_pos+=(sizeof(int)*2);
  }

//  fprintf(stderr,"\tAdded integer\n");
  return(0);
}

int code_addstr(tsm_code *tfc, const char *str)
{
  long int lastpos=tfc->mmap_pos+align_up(strlen(str)+sizeof(tsm_data_e)+1);
//  fprintf(stderr,"code_addstr(%p,\"%s\")\n",tfc,str);

  if(lastpos > tfc->mmap_end)
  {
    tsm_debug("\tCode seg full\n");
    return(-1);
  }

  *(tsm_data_e *)(tfc->mmap+tfc->mmap_pos) = TFD_STRING;
  strcpy(tfc->mmap + tfc->mmap_pos+sizeof(tsm_data_e),str);
  tfc->mmap_pos = lastpos;
//  fprintf(stderr,"\tAdded string\n");
  return(0);
}

int code_addinstr(tsm_code *tfc, tsm_inst_e instr)
{
//  fprintf(stderr,"code_addinstr(%p,\"%d\")\n",tfc,instr);

  if((tfc->mmap_pos+(sizeof(int)*2)) > tfc->mmap_end)
  {
    tsm_debug("\tCode seg full\n");
    return(-1);
  }

  {
    int *pos=(int *)(tfc->mmap + tfc->mmap_pos);
    pos[0]=TFD_INSTRUCTION;
    pos[1]=instr;
    tfc->mmap_pos+=(sizeof(int)*2);
  }

//  fprintf(stderr,"\tAdded instruction\n");
  return(0);
}
