/**
 * tsmforth.h -- defines functions for compilation and execution of
 * TSMforth code
 */
#ifndef __TSMFORTH_H__
#define __TSMFORTH_H__

#include "tsmtypes.h"
#include "tsmstack.h"
#include "strhash.h"
#include "debug.h"

/* Maximum number of function entry points for tsmforth code memory */
#define ENTRY_HASH (1024)
/* Maximum number of local variables */
#define LOCAL_MAX  (128)

/* What size the TSMforth stack should align itself on.  Certain
 * platforms may need specific alignments.  x86 will work with any 
 * alignment but faster for integer alignment.
 */
#define ALIGN_BIT sizeof(int)
#define ALIGN_MASK (ALIGN_BIT-1)
/* Maximum number of nested function calls. */
#define PROC_DEPTH 32

/* if-else-then/do-while branching stack.  Only used internally. */
typedef struct jumptab
{
  int pos;
  long table[1024];
} jumptab;

/* Dynamically-allocated codepage structure */
typedef struct tsm_code
{
  /* Base of memory map */
  unsigned char *mmap;
  /* Current position in writing to memory map */
  unsigned long  mmap_pos;
  /* End of memory map */
  unsigned long  mmap_end;
  /* Hashtable of function entry points */
  strtable      *entry;
  /* branching-stack for use in compilation */
  jumptab *aux;
} tsm_code;

/* Memory context for code execution */
typedef struct tsm_ctx
{
  /* Read-only codepage */
  tsm_code  *code;
  /* Data stack */
  tsm_stack *stack;
  /* Depth of function calls */
  int depth;
  /* Function call return points */
  unsigned long ip[PROC_DEPTH];
  /* Hashtable holding values of local variables */
  strtable *locals;
  /* If nonzero, will free the code page when context is freed */
  int freecode;
} tsm_ctx;

/* Structure describing a built-in instruction */
typedef struct tsm_handler
{
  /* Function to call on encountering instruction */
  int (*handler)(tsm_ctx *ctx);
  /* String representing instruction in source code */
  const char *token;
} tsm_handler;

#ifdef __cplusplus
extern "C" {
#endif

/* Array of built-in instructions */
extern const tsm_handler tsm_instr[];
/* valid instruction indexes are 0 <= n < TSM_LAST */
extern const int TSM_LAST;

/* Create a code-execution context from a code page and an entry point
 * name such as "main"
 */
tsm_ctx   *ctx_create(tsm_code *code, const char *entry);
/* Frees a code-execution context */
void      ctx_free(tsm_ctx *ctx);

/* Creates a codepage of given size */
tsm_code  *code_create(int size);
/* Frees codepage */
void      code_free(tsm_code *tfc);
/* Parses source file, adds it to the code page.  Returns NULL on error */
tsm_code *forth_parse(tsm_code *tfc, const char *fname);

/* These functions used internally by forth_parse(). */

/* Adds an entry point of current position to the code table */
int       code_addentry(tsm_code *tfc, const char *str);
/* Adds an immediate integer value to the codepage */
int       code_addint(tsm_code *tfc, int val);
/* Adds an immediate string value to the codepage */
int       code_addstr(tsm_code *tfc, const char *str);
/* Adds an instruction to the codepage */
int       code_addinstr(tsm_code *tfc, tsm_inst_e instr);
/* Returns the offset for a given entry point */
unsigned long code_getentry(tsm_code *tfc, const char *str);

#ifdef __cplusplus
}
#endif

/* Aligns stack addresses onto the given alignment if needed */
static __inline__ unsigned long align_up(unsigned long l)
{
  if(l&ALIGN_MASK)
  {
    l|=ALIGN_MASK;
    l++;
  }
  return(l);
}

/* Macro to push an integer onto a stack */
static __inline__ int tsm_stack_push_int(tsm_stack *s, const int i)
{
  return(tsm_stack_push(s,TFD_INT,&i));
}

/* Macro to push a string onto a stack */
static __inline__ int tsm_stack_push_string(tsm_stack *s, const char *str)
{
  return(tsm_stack_push(s,TFD_STRING,str));
}

#endif/*__TSMFORTH_H__*/
