/**
 * TSM PreProcessor.  Converts an input file stream into tokens, then
 * tsmforth instructions.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "strhash.h"
#include "tsmforth.h"
#include "debug.h"

#define PP_TAB_MAX 128
#define BR_TAB_MAX 32

static int pp_tab_init=0;

STATIC_HASHTABLE(pp_table,PP_TAB_MAX);
STATIC_HASHTABLE(br_table,BR_TAB_MAX);

int init_pp_table();

int str_is_int(const char *str)
{
  if(str[0]=='-')
    str++;

  if(!isdigit(str[0])) return(0);

  while(str[0]!='\0')
  {
    if(!isdigit(str[0])) return(0);
    str++;
  }

  return(1);
}

typedef enum fpp_state
{
  FPP_NONE,
  FPP_QUOTED,
  FPP_COMMENT,
  FPP_LVAR_PRE,
  FPP_LVAR,
  FPP_ERROR
}
fpp_state;

__inline__ void push_buf(char *buf, int *pos, int c) { buf[(*pos)++]=c; }

#define local_key(varname,filekey) (str_hash(varname)^(filekey))

int flush_buf(tsm_code *tfc, char *buf, int *pos, unsigned int filekey)
{
  int (*d)(tsm_code *tfc);

  if((*pos)==0) return(0);

  buf[(*pos)]='\0';

  DEBUG_PRINT(("0x%08x] FLUSH[%s]\n",tfc->mmap_pos,buf));

  d=hash_lookup_str(&br_table,buf);
  if(d!=NULL)
  {
    if(d(tfc)<0)
    {
      fprintf(stderr,"  Error processing branch '%s'\n",buf);
      return(-1);
    }
  }
  else switch(buf[0])
  {
  unsigned int k;
  case '@': // fetch
    DEBUG_PRINT(("\tFound fetch:  @%s\n",buf+1));
    if(code_addint(tfc,local_key(buf+1,filekey))<0)
      return(-1);
    if(code_addinstr(tfc,TSM_GET_LOCAL)<0)
      return(-1);    
    break;

  case '!': // store
    DEBUG_PRINT(("\tFound store:  !%s\n",buf+1));
    if(code_addint(tfc,local_key(buf+1,filekey))<0)
      return(-1);
    if(code_addinstr(tfc,TSM_SET_LOCAL)<0)
      return(-1);    
    break;

  case ';':
    DEBUG_PRINT(("\tFound ; (ret)\n"));

    if(buf[1]!='\0')
    {
      tsm_debug("Invalid return '%s'\n",buf);
      return(-1);
    }

    if(code_addinstr(tfc,TSM_RET)<0)
      return(-1);
    break;	

  case ':': // entry point
    DEBUG_PRINT(("\tFound entry point :%s\n",buf+1));
    if(code_addentry(tfc,buf+1)<0)
      return(-1);
    break;

  case '"': // string
    DEBUG_PRINT(("\tFound string %s\n",buf));
    if((*pos)<2)
    {
      tsm_debug("Invalid string\n");
      return(-1);
    }

    buf[(*pos)-1]='\0';
    if(code_addstr(tfc,buf+1)<0)
      return(-1);
    break;

  case '0': case '1': case '2': case '3': case '4': case '5':
  case '6': case '7': case '8': case '9': case '-':
    if(str_is_int(buf))
    {
      int i;
      DEBUG_PRINT(("\tFound integer %s\n",buf));
      i=atoi(buf);
      if(code_addint(tfc,i)<0)
        return(-1);

      break;
    }

    // NOTE -- falls through to default
  default:
    k=(int)(long int)hash_lookup_str(&pp_table,buf);
    if(k!=0)
    {
      DEBUG_PRINT(("\tFound instruction %d\n",k));
      DEBUG_PRINT(("\t(%s)\n",tsm_instr[k].token));
      if(code_addinstr(tfc,(long int)k)<0)
        return(-1);
    }
    else
    {
      k=str_hash(buf);
      DEBUG_PRINT(("\tAdding CALL %s\n",buf));
      if(code_addint(tfc,k)<0)
        return(-1);
      if(code_addinstr(tfc,TSM_CALL)<0)
        return(-1);
    }
    break;
  }

  *pos=0;
  buf[0]='\0';
  return(0);
}

tsm_code *forth_parse(tsm_code *tfc, const char *fname)
{
  char buf[2048];
  int bpos=0,curc=0,prevc=0;
  unsigned int filekey=str_hash(fname);
  fpp_state state=FPP_NONE;

  if(init_pp_table()<0)
  {
    tsm_debug("Couldn't init PP table\n");
    return(NULL);
  }

  FILE *fp=fopen(fname,"r");
  if(fp==NULL) return(NULL);

  tfc->aux=malloc(sizeof(jumptab));
  if(tfc->aux==NULL)
  {
    fclose(fp);
    return(NULL);
  }
  

  while(!feof(fp))
  {
    int escaped;
    prevc=curc; curc=fgetc(fp);
    if(curc<0) continue;
    escaped=(prevc=='\\');

    switch(state)
    {
/*
    case FPP_LVAR:
      if(isspace(curc))
      {
        if(flush_buf(tfc,buf,&bpos,filekey)<0)
        {
          tsm_debug("Couldn't flush [%s]\n",buf);
          return(NULL);
        }
        state=FPP_NONE;
        continue;
      }

      push_buf(buf,&bpos,curc);
      break;

    case FPP_LVAR_PRE:
      if(isspace(curc)) continue;
      push_buf(buf,&bpos,curc);
      state=FPP_LVAR;
      break;*/

    case FPP_NONE:
      switch(curc)
      {
      case ';': //case ':':
        if(flush_buf(tfc,buf,&bpos,filekey)<0)
        {
          tsm_debug("Couldn't flush [%s]\n",buf);
          return(NULL);
        }
        push_buf(buf,&bpos,curc);
        /* NOTE -- no break, falls through */
      case ' ': case '\r': case '\n': case '\t':
//        if(strcmp(buf,"LVAR")==0)
//        {
//          state=FPP_LVAR_PRE;
//          push_buf(buf,&bpos,' ');
//          continue;
//        }

        if(flush_buf(tfc,buf,&bpos,filekey)<0)
        {
          tsm_debug("Couldn't flush [%s]\n",buf);
          return(NULL);
        }
        continue;
      case '"':
        if(flush_buf(tfc,buf,&bpos,filekey)<0)
        {
          tsm_debug("Couldn't flush [%s]\n",buf);
          return(NULL);
        }
        state=FPP_QUOTED;
        break;
      case '(':
        if(flush_buf(tfc,buf,&bpos,filekey)<0)
        {
          tsm_debug("Couldn't flush [%s]\n",buf);
          return(NULL);
        }
        state=FPP_COMMENT;
        continue;
      case ')':
        return(NULL);
      }

      push_buf(buf,&bpos,curc);
      break;

    case FPP_QUOTED:
      if(prevc=='\\')
      {
        switch(curc)
        {
        case 'a':  push_buf(buf,&bpos,'\a'); break;
        case 'b':  push_buf(buf,&bpos,'\b'); break;
        case 'r':  push_buf(buf,&bpos,'\r'); break;
        case 'n':  push_buf(buf,&bpos,'\n'); break;
        case 't':  push_buf(buf,&bpos,'\t'); break;
        case 'f':  push_buf(buf,&bpos,'\f'); break;
        case '\\': push_buf(buf,&bpos,'\\'); break;
        case 'v':  push_buf(buf,&bpos,'\v'); break;
        default:   push_buf(buf,&bpos,curc); break;
        }
        break;
      }

      if(curc=='\\') break;
      push_buf(buf,&bpos,curc);
      if(curc=='"')
      {
        if(flush_buf(tfc,buf,&bpos,filekey)<0)
        {
          tsm_debug("Couldn't flush [%s]\n",buf);
          return(NULL);
        }
        state=FPP_NONE;
      }
      break;
    case FPP_COMMENT:
      if(curc==')') state=FPP_NONE;
      break;
    case FPP_ERROR:
      return(NULL);
    }
  }  

  if(flush_buf(tfc,buf,&bpos,filekey)<0)
  {
    tsm_debug("Couldn't flush [%s]\n",buf);
    return(NULL);
  }

  free(tfc->aux);
  tfc->aux=NULL;

  return(tfc);
}

#ifdef TSMPRE_TEST
int main(int argc, char *argv[])
{
  tsm_code *tfc=code_create(1024*1024*256);
  fprintf(stderr,"tfc=%p\n",tfc);
  forth_parse(tfc,argv[1]);
  fprintf(stderr,"tfc=%p\n",tfc);
  code_free(tfc);
  return(0);
}
#endif/*TSMPRE_TEST*/

static int tfpp_handle_while(tsm_code *tfc)
{
  /* retrieve offset from table */    
  long off=tfc->aux->table[--(tfc->aux->pos)];
  long jmpoff=off-(tfc->mmap_pos)-(sizeof(tsm_codedata)*2);

  /* record offset in table */
  if(code_addint(tfc,jmpoff)<0)
    return(-1);

  /* record instruction in code table */
  if(code_addinstr(tfc,TSM_IFN)<0)
    return(-1);

  return(0);
}

static int tfpp_handle_if(tsm_code *tfc)
{
  /* push offset in aux table */
  tfc->aux->table[tfc->aux->pos++]=tfc->mmap_pos;

  /* record NULL in code table */
  if(code_addint(tfc,0)<0)
    return(-1);
  /* record instruction in code table */
  if(code_addinstr(tfc,TSM_IF)<0)
    return(-1);
  return(0);
}

static int tfpp_handle_then(tsm_code *tfc)
{
  /* retrieve offset from table */    
  long off=tfc->aux->table[--(tfc->aux->pos)];
  /* update offset in code table */
  tsm_codedata *cd=(tsm_codedata *)(tfc->mmap+off);

  if(cd->type!=TFD_INT)
  {
    tsm_debug("\tError setting up jump, does not point to int\n");
    return(-1);
  }
  if(cd->value.intvalue != 0)
  {
    tsm_debug("\tError -- no jump placeholder\n");
    return(-1);
  }
  cd->value.intvalue=(tfc->mmap_pos-off)-(sizeof(tsm_codedata)*2);
  /* do NOT record instruction! */
  return(0);
}

static int tfpp_handle_do(tsm_code *tfc)
{
  tfc->aux->table[tfc->aux->pos++]=tfc->mmap_pos;    
  return(0);  
}

static int tfpp_handle_else(tsm_code *tfc)
{
  /* Get offset from jump table */
  long int off=tfc->aux->table[--tfc->aux->pos];
  tsm_codedata *cd;

  /* Record offset in jump table */
  tfc->aux->table[tfc->aux->pos++]=tfc->mmap_pos;

  /* record NULL in code table */
  if(code_addint(tfc,0)<0)
    return(-1);
  /* record instruction in code table */
  if(code_addinstr(tfc,TSM_JUMP)<0)
    return(-1);
  /* Record current offset in code table */
  cd=(tsm_codedata *)(tfc->mmap+off);

  if(cd->type!=TFD_INT)
  {
    tsm_debug("\tError setting up jump, does not point to int\n");
    return(-1);
  }
  if(cd->value.intvalue != 0)
  {
    tsm_debug("\tError -- no jump placeholder\n");
    return(-1);
  }

  cd->value.intvalue=(tfc->mmap_pos-off)-(sizeof(tsm_codedata)*2);
  return(0);
}

typedef struct br_handler
{
  const char *token;
  int (*func)(tsm_code *);
} br_handler;

static const br_handler br_handlers[]={
  {DO_STR   ,tfpp_handle_do   }
, {WHILE_STR,tfpp_handle_while}
, {IF_STR   ,tfpp_handle_if   }
, {THEN_STR ,tfpp_handle_then }
, {ELSE_STR ,tfpp_handle_else }
, {NULL     ,NULL             }
};

int init_pp_table()
{
  int n;
  if(pp_tab_init) return(0);

  STATIC_HASHINIT(pp_table);
  STATIC_HASHINIT(br_table);

  for(n=TSM_BUILTIN_MAX; tsm_instr[n].token!=NULL; n++)
    if(hash_add_str_static(&pp_table,tsm_instr[n].token,(void *)(long int)n)<0)
    {
      tsm_debug("Preprocessor collision \"%s\"!\n"
              "Make table larger or fix hash func!\n",
              tsm_instr[n].token);
      return(-1);
    }

  for(n=0; br_handlers[n].token!=NULL; n++)
    if(hash_add_str_static(&br_table,br_handlers[n].token,
      (void *)br_handlers[n].func)<0)
    {
      tsm_debug("break-table collision \"%s\"!\n"
              "Make table larger or fix hash func!\n",
              tsm_instr[n].token);
      return(-1);
    }

  pp_tab_init=1;

  return(0);
}

